from __future__ import annotations

import shutil
import subprocess
from concurrent.futures import ThreadPoolExecutor, as_completed
from hashlib import sha1
from pathlib import Path
from typing import Any, Dict, NamedTuple

RUFF_BASE_ARGS = [
    "ruff",
    "--no-cache",
    "--fix",
    "--quiet",
    "--select",
    "UP",
    "--select",
    "F401",
    "--isolated",
    "-",
    "--target-version",
]


class VersionTuple(NamedTuple):
    major: int
    minor: int

    @classmethod
    def from_string(cls, version: str) -> VersionTuple:
        major, minor = version.split(".")
        return VersionTuple(major=int(major), minor=int(minor))


VersionedCode = Dict[VersionTuple, str]


class Cache:
    """Simple hybrid file system / memory cache.

    Follows the
    `Cache Directory Tagging Specification http://www.brynosaurus.com/cachedir/>`_.
    """

    def __init__(self) -> None:
        self.cache_dir = Path(".auto_pytabs_cache")
        self.cache_content_dir = self.cache_dir / "content"
        self._cache: dict[str, str] = {}
        self._touched: set[str] = set()
        self._load()

    def _init_cache_dir(self) -> None:
        self.cache_content_dir.mkdir(exist_ok=True, parents=True)
        cachedir_tag = self.cache_dir / "CACHEDIR.TAG"
        gitignore_file = self.cache_dir / ".gitignore"
        if not cachedir_tag.exists():
            cachedir_tag.write_text(
                """Signature: 8a477f597d28d172789f06886806bc55
# This file is a cache directory tag created by (application name).
# For information about cache directory tags, see:
#	http://www.brynosaurus.com/cachedir/"""
            )
        if not gitignore_file.exists():
            gitignore_file.write_text("*\n")

    def _load(self) -> None:
        self._init_cache_dir()

        cache: dict[str, str] = {}
        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(file.read_text): file.name
                for file in self.cache_content_dir.iterdir()
            }
            for future in as_completed(futures):
                cache[futures[future]] = future.result()
        self._cache.update(cache)

    @staticmethod
    def make_cache_key(*parts: Any) -> str:
        """Create a cache key using an md5 hash of ``parts``."""
        return sha1("".join(map(str, parts)).encode()).hexdigest()

    def get(self, key: str) -> str | None:
        """Get an item specified by ``key`` the cache."""
        self._touched.add(key)
        return self._cache.get(key)

    def set(self, key: str, content: str) -> None:
        """Store an ``content``."""
        self._cache[key] = content
        self._touched.add(key)

    def persist(self, evict: bool = True) -> None:
        """
        Persist internal cache to disk. If ``evict`` is ``True``, evict unused items.
        """
        with ThreadPoolExecutor() as executor:
            for key, content in self._cache.items():
                if key in self._touched:
                    executor.submit(
                        self.cache_content_dir.joinpath(key).write_text, content
                    )
                elif evict:
                    executor.submit(
                        self.cache_content_dir.joinpath(key).unlink, missing_ok=True
                    )

    def clear_all(self) -> None:
        """Clear all cached items from memory and disk."""
        self._cache = {}
        self._touched = set()

        if not self.cache_dir.exists():
            return

        shutil.rmtree(self.cache_dir)


def get_version_requirements(
    min_version: VersionTuple, max_version: VersionTuple
) -> list[VersionTuple]:
    """Given a min and max version, generate all versions in between."""
    min_major, min_minor = min_version
    max_major, max_minor = max_version
    return [
        VersionTuple(major=major, minor=minor)
        for major in range(min_major, max_major + 1)
        for minor in range(min_minor, max_minor + 1)
    ]


def _run_ruff(source: str, target_version: str) -> str:
    with subprocess.Popen(
        [*RUFF_BASE_ARGS, target_version],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        encoding="utf-8",
    ) as process:
        return process.communicate(input=source)[0]


def _upgrade_code(
    code: str, min_version: VersionTuple, cache: Cache | None = None
) -> str:
    cache_key: str | None = None
    if cache:
        cache_key = cache.make_cache_key(code, min_version)

        if cached_code := cache.get(cache_key):
            return cached_code

    code = _run_ruff(code, target_version=f"py3{min_version.minor}")

    if cache_key and cache:
        cache.set(cache_key, code)

    return code


def version_code(
    code: str, version_requirements: list[VersionTuple], cache: Cache | None = None
) -> VersionedCode:
    """Create versions of ``code`` for all python versions specified in
    ``version_requirements`` and return a dictionary of version-tuples/code.
    """
    latest_code = code
    versioned_code: VersionedCode = {version_requirements[0]: code}

    for version in version_requirements:
        upgraded_code = _upgrade_code(latest_code, version, cache=cache)
        if upgraded_code != latest_code:
            versioned_code[version] = upgraded_code
            latest_code = upgraded_code

    return versioned_code



from __future__ import annotations

from pathlib import Path
from test.conftest import SphinxBuilder
from typing import TYPE_CHECKING, Callable

import pytest

if TYPE_CHECKING:
    from pytest_regressions.file_regression import FileRegressionFixture

SphinxBuilderFixture = Callable[..., SphinxBuilder]

TEST_DATA_DIR = Path("test/sphinx_ext_test_data")
SOURCE_CONTENT = (TEST_DATA_DIR / "test.rst").read_text()
SOURCE_CONTENT_COMPAT = (TEST_DATA_DIR / "test_compat.rst").read_text()


@pytest.fixture(params=[True, False], ids=lambda item: f"compat:{item}")
def compat(request) -> bool:
    return request.param


@pytest.fixture
def source(compat: bool) -> str:
    if compat:
        return SOURCE_CONTENT_COMPAT
    return SOURCE_CONTENT


@pytest.mark.parametrize("no_cache", [True, False], ids=lambda item: f"nocache:{item}")
def test_upgrade_tabs_all_versions(
    sphinx_builder: SphinxBuilderFixture,
    file_regression: FileRegressionFixture,
    no_cache: bool,
    compat: bool,
    source: str,
    mock_cache_persist,
):
    builder = sphinx_builder(
        source=source, auto_pytabs_no_cache=no_cache, compat=compat
    )

    builder.build()

    pformat = builder.get_doctree("index").pformat()
    file_regression.check(pformat, fullpath=TEST_DATA_DIR / "tabs_all_versions.xml")

    if no_cache:
        mock_cache_persist.assert_not_called()
    else:
        mock_cache_persist.assert_called_once_with()


def test_upgrade_reverse_order(
    sphinx_builder: SphinxBuilderFixture,
    file_regression: FileRegressionFixture,
    compat: bool,
    source: str,
):
    builder = sphinx_builder(
        source=source,
        auto_pytabs_no_cache=True,
        auto_pytabs_reverse_order=True,
        compat=compat,
    )

    builder.build()

    file_regression.force_regen = True
    pformat = builder.get_doctree("index").pformat()
    file_regression.check(
        pformat, fullpath=TEST_DATA_DIR / "tabs_all_versions_reversed.xml"
    )


def test_upgrade_single_version(
    sphinx_builder: SphinxBuilderFixture,
    file_regression: FileRegressionFixture,
    compat: bool,
    source: str,
):
    builder = sphinx_builder(
        source=source, compat=compat, auto_pytabs_min_version=(3, 11)
    )

    builder.build()

    pformat = builder.get_doctree("index").pformat()
    file_regression.check(pformat, fullpath=TEST_DATA_DIR / "tabs_single_version.xml")


@pytest.mark.parametrize(
    "min_version,max_version",
    [
        pytest.param((3, i), (3, j), id=f"min:3.{i}-max:3.{j}")
        for i in range(8, 12)
        for j in range(7, 12)
        if i <= j
    ],
)
def test_upgrade_versions(
    min_version: tuple[int, int],
    max_version: tuple[int, int],
    sphinx_builder: SphinxBuilderFixture,
    file_regression: FileRegressionFixture,
    compat: bool,
    source: str,
) -> None:
    if min_version > max_version:
        pytest.skip()
    builder = sphinx_builder(
        source=source,
        compat=compat,
        auto_pytabs_min_version=min_version,
        auto_pytabs_max_version=max_version,
    )

    builder.build()

    pformat = builder.get_doctree("index").pformat()

    file_regression.check(
        pformat,
        fullpath=TEST_DATA_DIR.joinpath(
            f"tabs_versions_min-{min_version[0]}{min_version[1]}_max-{max_version[0]}{max_version[1]}.xml"
        ),
    )


@pytest.mark.parametrize("reverse_order", [True, False])
@pytest.mark.parametrize("default_tab_version", ["highest", "lowest"])
def test_upgrade_default_tab(
    default_tab_version: str,
    sphinx_builder: SphinxBuilderFixture,
    file_regression: FileRegressionFixture,
    source: str,
    compat: bool,
    reverse_order: bool,
) -> None:
    builder = sphinx_builder(
        source=source,
        compat=compat,
        auto_pytabs_default_tab=default_tab_version,
        auto_pytabs_min_version=(3, 7),
        auto_pytabs_max_version=(3, 10),
        auto_pytabs_reverse_order=reverse_order,
    )
    builder.build()

    pformat = builder.get_doctree("index").pformat()

    file_stem = f"tabs_default_tab_{default_tab_version}"
    if reverse_order:
        file_stem += "_reversed"

    file_regression.check(pformat, fullpath=TEST_DATA_DIR.joinpath(file_stem + ".xml"))



import pytest
from auto_pytabs.core import Cache


@pytest.fixture()
def cache() -> Cache:
    cache = Cache()
    cache.clear_all()
    cache._init_cache_dir()
    yield cache
    cache.clear_all()


def test_get_not_found(cache) -> None:
    assert cache.get("foo") is None


def test_get(cache) -> None:
    cache.cache_content_dir.joinpath("foo").write_text("bar")
    cache._load()

    assert cache.get("foo") == "bar"
    assert cache._cache["foo"] == "bar"


def test_set(cache) -> None:
    cache.set("foo", "bar")

    assert cache.get("foo") == "bar"


@pytest.mark.parametrize("evict", [True, False])
def test_persist(cache: Cache, evict: bool) -> None:
    test_file_one = cache.cache_content_dir.joinpath("one")
    test_file_two = cache.cache_content_dir.joinpath("two")
    test_file_one.write_text("foo")
    test_file_two.write_text("bar")
    cache._load()

    cache.get("two")
    cache.set("three", "baz")

    cache.persist(evict=evict)

    if evict:
        assert not test_file_one.exists()
    else:
        assert test_file_one.exists()
    assert cache.cache_content_dir.joinpath("two").exists()
    assert cache.cache_content_dir.joinpath("three").exists()


def test_clear_all(cache) -> None:
    cache.set("foo", "bar")

    cache.clear_all()

    assert not cache.cache_dir.exists()



from __future__ import annotations

from typing import Literal

import pytest
import yaml
from auto_pytabs.core import Cache
from auto_pytabs.mkdocs_plugin import AutoPyTabsPlugin
from mkdocs.config.base import Config, load_config


@pytest.fixture()
def setup_mkdocs(tmp_path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    docs_dir = tmp_path / "docs"
    docs_dir.mkdir(exist_ok=True)
    docs_dir.joinpath("index.md").write_text("# Test\n")


@pytest.fixture()
def create_mkdocs_config(request, tmp_path, setup_mkdocs):
    def _create_mkdocs_config(
        *,
        min_version: str | None = None,
        max_version: str | None = None,
        tab_title_template: str | None = None,
        no_cache: bool = False,
        default_tab: Literal["highest", "lowest"] | None = None,
        reverse_order: bool | None = None,
    ) -> Config:
        auto_pytabs_config = {}
        if min_version:
            auto_pytabs_config["min_version"] = min_version
        if max_version:
            auto_pytabs_config["max_version"] = max_version
        if tab_title_template:
            auto_pytabs_config["tab_title_template"] = tab_title_template
        if no_cache:
            auto_pytabs_config["no_cache"] = no_cache
        if default_tab:
            auto_pytabs_config["default_tab"] = default_tab
        if reverse_order is not None:
            auto_pytabs_config["reverse_order"] = reverse_order

        config = {
            "site_name": "Test docs",
            "markdown_extensions": [{"pymdownx.tabbed": None}],
            "plugins": [{"auto_pytabs": auto_pytabs_config}],
        }

        config_file = tmp_path / "mkdocs.yaml"
        config_file.write_text(yaml.dump(config))

        request.addfinalizer(config_file.unlink)

        return load_config(str(config_file))

    return _create_mkdocs_config


@pytest.fixture()
def configured_plugin(create_mkdocs_config) -> tuple[AutoPyTabsPlugin, Config]:
    config = create_mkdocs_config()
    plugin = config["plugins"]["auto_pytabs"]
    plugin.on_config(config)
    return plugin, config


@pytest.mark.parametrize(
    "reverse_order,expected_reverse_order_value",
    [(None, False), (True, True), (False, False)],
)
@pytest.mark.parametrize(
    "default_tab,expected_default_tab_value",
    [(None, "highest"), ("highest", "highest"), ("lowest", "lowest")],
)
@pytest.mark.parametrize("no_cache", [True, False])
@pytest.mark.parametrize(
    "tab_title_template", ("{min_version}", "Python {min_version} and higher")
)
@pytest.mark.parametrize("min_version,max_version", [("3.7", "3.11"), ("3.8", "3.10")])
def test_config(
    min_version,
    max_version,
    no_cache,
    tab_title_template,
    create_mkdocs_config,
    default_tab: str | None,
    expected_default_tab_value: str,
    reverse_order: bool | None,
    expected_reverse_order_value: bool,
) -> None:
    config = create_mkdocs_config(
        min_version=min_version,
        max_version=max_version,
        tab_title_template=tab_title_template,
        no_cache=no_cache,
        default_tab=default_tab,
        reverse_order=reverse_order,
    )
    plugin = config["plugins"]["auto_pytabs"]

    assert plugin
    assert isinstance(plugin, AutoPyTabsPlugin)

    plugin.on_config(config)

    received_config = config["mdx_configs"]["auto_pytabs"]

    assert "auto_pytabs" in config["markdown_extensions"]
    assert received_config["min_version"] == min_version
    assert received_config["max_version"] == max_version
    assert received_config["tab_title_template"] == tab_title_template
    assert received_config["cache"] is plugin.cache
    assert received_config["default_tab"] == expected_default_tab_value
    assert received_config["reverse_order"] == expected_reverse_order_value

    if no_cache:
        assert plugin.cache is None
    else:
        assert isinstance(plugin.cache, Cache)


def test_on_post_build(configured_plugin, mock_cache_persist) -> None:
    plugin, config = configured_plugin

    plugin.on_post_build(config)

    mock_cache_persist.assert_called_once_with()


def test_on_build_error(configured_plugin, mock_cache_persist) -> None:
    plugin, config = configured_plugin

    plugin.on_build_error(ValueError())

    mock_cache_persist.assert_called_once_with(evict=False)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from auto_pytabs.markdown_ext import UpgradePreprocessor

if TYPE_CHECKING:
    from pytest_regressions.file_regression import FileRegressionFixture

TEST_DATA_DIR = Path("test/md_ext_test_data")


@pytest.fixture()
def preprocessor() -> UpgradePreprocessor:
    return UpgradePreprocessor(
        min_version="3.7", max_version="3.11", default_tab_strategy="highest"
    )


def get_test_data(name: str) -> tuple[list[str], Path]:
    in_file = TEST_DATA_DIR / f"{name}_in.md"
    out_file = TEST_DATA_DIR / f"{name}_out.md"
    return in_file.read_text().splitlines(), out_file


def test_upgrade_single_version(file_regression: FileRegressionFixture):
    source, expected_output_path = get_test_data("upgrade_single")
    preprocessor = UpgradePreprocessor(
        min_version="3.9", max_version="3.11", default_tab_strategy="highest"
    )

    output = "\n".join(preprocessor.run(source))
    file_regression.check(output, fullpath=expected_output_path)


@pytest.mark.parametrize(
    "reverse_order", [False, True], ids=lambda item: f"reverse:{item}"
)
@pytest.mark.parametrize("default_tab_strategy", ["highest", "lowest"])
def test_upgrade(
    default_tab_strategy: str,
    reverse_order: bool,
    file_regression: FileRegressionFixture,
):
    source = TEST_DATA_DIR.joinpath("upgrade_in.md").read_text()
    file_stem = f"upgrade_out_default_{default_tab_strategy}"
    if reverse_order:
        file_stem += "_reversed"
    expected_output = TEST_DATA_DIR.joinpath(file_stem + ".md")

    preprocessor = UpgradePreprocessor(
        min_version="3.7",
        max_version="3.11",
        default_tab_strategy=default_tab_strategy,
        reverse_order=reverse_order,
    )
    file_regression.force_regen = True
    output = "\n".join(preprocessor.run(source.splitlines()))
    file_regression.check(output, fullpath=expected_output)


def test_upgrade_custom_tab_title(file_regression: FileRegressionFixture):
    preprocessor = UpgradePreprocessor(
        min_version="3.7",
        max_version="3.11",
        tab_title_template="Python {min_version} and above",
        default_tab_strategy="highest",
    )
    source, expected_output_path = get_test_data("custom_tab_title")

    output = "\n".join(preprocessor.run(source))
    file_regression.check(output, fullpath=expected_output_path)


def test_nested_tabs(
    preprocessor: UpgradePreprocessor, file_regression: FileRegressionFixture
):
    source, expected_output_path = get_test_data("nested_tabs")

    output = "\n".join(preprocessor.run(source))
    file_regression.check(output, fullpath=expected_output_path)


def test_disable_block(
    preprocessor: UpgradePreprocessor, file_regression: FileRegressionFixture
):
    source, expected_output_path = get_test_data("disable_block")

    output = "\n".join(preprocessor.run(source))
    file_regression.check(output, fullpath=expected_output_path)


def test_disable_section(
    preprocessor: UpgradePreprocessor, file_regression: FileRegressionFixture
):
    source, expected_output_path = get_test_data("disable_section")

    output = "\n".join(preprocessor.run(source))
    file_regression.check(output, fullpath=expected_output_path)


def test_ignore_fenced_block(preprocessor: UpgradePreprocessor):
    source = """```json
{"foo": "bar"}
```"""

    output = "\n".join(preprocessor.run(source.splitlines()))
    assert output == source



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mkdocs.config.defaults import MkDocsConfig


from mkdocs.config import Config, config_options
from mkdocs.plugins import BasePlugin

from auto_pytabs.core import Cache


class PluginConfig(Config):  # type: ignore[no-untyped-call]
    min_version = config_options.Type(str, default="3.7")
    max_version = config_options.Type(str, default="3.11")
    tab_title_template = config_options.Type(str, default="Python {min_version}+")
    no_cache = config_options.Type(bool, default=False)
    default_tab = config_options.Choice(["highest", "lowest"], default="highest")
    reverse_order = config_options.Type(bool, default=False)


class AutoPyTabsPlugin(BasePlugin[PluginConfig]):  # type: ignore[no-untyped-call]
    def __init__(self) -> None:
        self.cache: Cache | None = None

    def on_config(self, config: MkDocsConfig) -> Config | None:
        if not self.config.no_cache:
            self.cache = Cache()

        config.markdown_extensions.append("auto_pytabs")
        config["mdx_configs"].update(
            {
                "auto_pytabs": {
                    "min_version": self.config.min_version,
                    "max_version": self.config.max_version,
                    "tab_title_template": self.config.tab_title_template,
                    "cache": self.cache,
                    "default_tab": self.config.default_tab,
                    "reverse_order": self.config.reverse_order,
                }
            }
        )
        return None

    def on_post_build(self, config: MkDocsConfig) -> None:
        if self.cache:
            self.cache.persist()

    def on_build_error(self, error: Exception) -> None:
        if self.cache:
            self.cache.persist(evict=False)



"""Largely adopted from
https://github.com/executablebooks/sphinx-design/blob/6df47513e9e221c61877e9308da7a41d216ae3c3/tests/conftest.py.
"""

from __future__ import annotations

import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest
from auto_pytabs.core import Cache
from sphinx.testing.path import path as sphinx_path

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from docutils import nodes
    from sphinx.testing.util import SphinxTestApp


pytest_plugins = "sphinx.testing.fixtures"


@pytest.fixture(autouse=True, scope="session")
def purge_cache():
    Cache().clear_all()
    yield
    Cache().clear_all()


@pytest.fixture()
def mock_cache_persist(mocker) -> MagicMock:
    return mocker.patch("auto_pytabs.core.Cache.persist")


class SphinxBuilder:
    def __init__(self, app: SphinxTestApp, src_path: Path):
        self.app = app
        self._src_path = src_path

    @property
    def src_path(self) -> Path:
        return self._src_path

    @property
    def out_path(self) -> Path:
        return Path(self.app.outdir)

    def build(self, assert_pass=True):
        self.app.build()
        if assert_pass:
            assert self.warnings == "", self.status
        return self

    @property
    def status(self):
        return self.app._status.getvalue()

    @property
    def warnings(self):
        return self.app._warning.getvalue()

    def get_doctree(
        self, docname: str, post_transforms: bool = False
    ) -> nodes.document:
        doctree: nodes.document = self.app.env.get_doctree(docname)
        if post_transforms:
            self.app.env.apply_post_transforms(doctree, docname)
        # make source path consistent for test comparisons
        for node in doctree.findall(include_self=True):
            if not ("source" in node and node["source"]):
                continue
            node["source"] = Path(node["source"]).relative_to(self.src_path).as_posix()
            if node["source"].endswith(".rst"):
                node["source"] = node["source"][:-4]
            elif node["source"].endswith(".md"):
                node["source"] = node["source"][:-3]
        return doctree


@pytest.fixture()
def sphinx_builder(tmp_path: Path, make_app, monkeypatch):
    def _create_project(
        source: str,
        compat: bool = False,
        **conf_kwargs: dict[str, Any],
    ):
        if compat:
            conf_kwargs["auto_pytabs_compat_mode"] = True
        src_path = tmp_path / "srcdir"
        src_path.mkdir()
        conf_kwargs = {
            "extensions": [
                "sphinx_design",
                "auto_pytabs.sphinx_ext",
            ],
            "auto_pytabs_no_cache": True,
            **(conf_kwargs or {}),
        }
        content = "\n".join(
            [f"{key} = {value!r}" for key, value in conf_kwargs.items()]
        )
        src_path.joinpath("conf.py").write_text(content, encoding="utf8")
        app = make_app(srcdir=sphinx_path(str(src_path.resolve())), buildername="html")
        shutil.copy(
            "test/sphinx_ext_test_data/example.py", src_path.joinpath("example.py")
        )
        shutil.copy(
            "test/sphinx_ext_test_data/example.js", src_path.joinpath("example.js")
        )

        src_path.joinpath("index.rst").write_text(source)
        return SphinxBuilder(app, src_path)

    yield _create_project



from __future__ import annotations

import importlib
import importlib.metadata
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal, cast

from docutils.nodes import Node, container, section
from docutils.parsers.rst import directives
from docutils.statemachine import ViewList
from sphinx.directives.code import CodeBlock, LiteralInclude
from sphinx.util.docutils import SphinxDirective
from sphinx.util.nodes import nested_parse_with_titles

from auto_pytabs.core import (
    Cache,
    VersionedCode,
    get_version_requirements,
    version_code,
)

if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.config import Config
    from sphinx.environment import BuildEnvironment


def indent(
    text: str | Iterable[str], indent_char: str = " ", level: int = 4
) -> list[str]:
    lines = text.splitlines() if isinstance(text, str) else text
    return list((indent_char * level) + line for line in lines)


def _render_directive(
    *,
    name: str,
    argument: str = "",
    options: dict[str, Any] | None = None,
    body: str | Iterable[str],
) -> list[str]:
    directive = [f".. {name}:: {argument}"]
    if options:
        rendered_options = [
            f":{option}: {value if (value is not True and value) else ''}"
            for option, value in options.items()
            if value is not False
        ]
        directive.extend(indent(rendered_options))

    directive.append("")
    directive.extend(indent(body))

    return directive


class UpgradeMixin(SphinxDirective):
    compat: bool = False

    def _get_directive_options(self) -> dict[str, Any]:
        options: dict[str, Any] = {}
        if not self.compat:
            options["no-upgrade"] = True
        for option, value in self.options.items():
            if option not in CodeBlock.option_spec:
                continue
            if self.option_spec[option] is directives.flag:
                value = True
            if isinstance(value, Iterable) and not isinstance(value, str):
                value = "\n".join(value)
            options[option] = value
        return options

    def _create_tabs(
        self,
        versioned_code: VersionedCode,
        tab_title_template: str,
    ) -> list[str]:
        directive_options = self._get_directive_options()
        if len(versioned_code) == 1:
            return _render_directive(
                name="code-block",
                argument="python",
                body=versioned_code.popitem()[1],
                options=directive_options,
            )

        default_tab_strategy: Literal["highest", "lowest"] = self.config[
            "auto_pytabs_default_tab"
        ]
        versions = list(versioned_code.keys())
        default_selected_version = versions[
            -1 if default_tab_strategy == "highest" else 0
        ]

        tab_set_body = []
        if self.config["auto_pytabs_reverse_order"]:
            versioned_code = dict(reversed(versioned_code.items()))

        for version, code in versioned_code.items():
            version_string = f"{version[0]}.{version[1]}"
            code_block = _render_directive(
                name="code-block",
                argument="python",
                options=directive_options,
                body=code,
            )
            tab_item = _render_directive(
                name="tab-item",
                argument=tab_title_template.format(min_version=version_string),
                options={
                    "sync": version_string,
                    "selected": version == default_selected_version,
                },
                body=code_block,
            )
            tab_set_body.extend(tab_item)

        return _render_directive(name="tab-set", body=tab_set_body)

    @property
    def cache(self) -> Cache | None:
        return getattr(self.env, "auto_pytabs_cache", None)

    def _create_py_tab_nodes(self, code: str) -> list[Node]:
        version_requirements = self.config["auto_pytabs_versions"]
        versioned_code = version_code(code, version_requirements, cache=self.cache)
        tabs = self._create_tabs(
            versioned_code, self.env.config["auto_pytabs_tab_title_template"]
        )

        rst = ViewList()
        source, lineno = self.get_source_info()
        for line in tabs:
            rst.append(line, source, lineno)

        node = section()
        node.document = self.state.document

        nested_parse_with_titles(self.state, rst, node)
        nodes = node.children

        return cast("list[Node]", nodes)


class PyTabsCodeBlock(CodeBlock, UpgradeMixin):
    compat = True

    def run(self) -> list[Node]:
        if not self.arguments or self.arguments[0] != "python":
            return super().run()

        self.assert_has_content()
        return self._create_py_tab_nodes("\n".join(self.content))


class PyTabsLiteralInclude(LiteralInclude, UpgradeMixin):
    compat = True

    def run(self) -> list[Node]:
        base_node = super().run()[0]
        if self.options.get("language") != "python":
            return [base_node]
        if isinstance(base_node, container):
            base_node = base_node.children[1]
        return self._create_py_tab_nodes(base_node.rawsource)  # type: ignore[attr-defined]  # noqa: E501


class CodeBlockOverride(PyTabsCodeBlock):
    compat = False
    option_spec = {**CodeBlock.option_spec, "no-upgrade": directives.flag}

    def run(self) -> list[Node]:
        if "no-upgrade" in self.options:
            return CodeBlock.run(self)

        return super().run()


class LiteralIncludeOverride(PyTabsLiteralInclude):
    compat = False
    option_spec = {**LiteralInclude.option_spec, "no-upgrade": directives.flag}

    def run(self) -> list[Node]:
        if "no-upgrade" in self.options:
            return LiteralInclude.run(self)
        return super().run()


def on_config_inited(app: Sphinx, config: Config) -> None:
    config["auto_pytabs_versions"] = get_version_requirements(
        config["auto_pytabs_min_version"], config["auto_pytabs_max_version"]
    )

    if not config["auto_pytabs_compat_mode"]:
        app.add_directive("code-block", CodeBlockOverride, override=True)
        app.add_directive("literalinclude", LiteralIncludeOverride, override=True)


def on_build_finished(app: Sphinx, exception: Exception | None) -> None:
    if cache := getattr(app.env, "auto_pytabs_cache", None):
        cache.persist()


def on_env_before_read_docs(
    app: Sphinx, env: BuildEnvironment, docnames: list[str]
) -> None:
    if not app.config["auto_pytabs_no_cache"]:
        env.auto_pytabs_cache = Cache()  # type: ignore[attr-defined]


def on_env_merge_info(
    app: Sphinx, env: BuildEnvironment, docnames: list[str], other: BuildEnvironment
) -> None:
    cache: Cache | None = getattr(env, "auto_pytabs_cache", None)
    other_cache: Cache | None = getattr(other, "auto_pytabs_cache", None)
    if cache and other_cache:
        cache._touched.update(other_cache._touched)
        cache._cache.update(other_cache._cache)


def setup(app: Sphinx) -> dict[str, bool | str]:
    app.add_directive("pytabs-code-block", PyTabsCodeBlock)
    app.add_directive("pytabs-literalinclude", PyTabsLiteralInclude)

    app.add_config_value(
        "auto_pytabs_tab_title_template",
        default="Python {min_version}+",
        rebuild="html",
    )
    app.add_config_value("auto_pytabs_min_version", default=(3, 7), rebuild="html")
    app.add_config_value("auto_pytabs_max_version", default=(3, 11), rebuild="html")
    app.add_config_value("auto_pytabs_no_cache", default=False, rebuild="html")
    app.add_config_value("auto_pytabs_compat_mode", default=False, rebuild="html")
    app.add_config_value("auto_pytabs_default_tab", default="highest", rebuild="html")
    app.add_config_value("auto_pytabs_reverse_order", default=False, rebuild="html")

    app.connect("config-inited", on_config_inited)
    app.connect("build-finished", on_build_finished)
    app.connect("env-before-read-docs", on_env_before_read_docs)
    app.connect("env-merge-info", on_env_merge_info)

    return {
        "version": importlib.metadata.version("auto_pytabs"),
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Literal, cast

from markdown import Extension
from markdown.preprocessors import Preprocessor

from auto_pytabs.core import (
    Cache,
    VersionedCode,
    VersionTuple,
    get_version_requirements,
    version_code,
)

if TYPE_CHECKING:
    from markdown import Markdown


RGX_BLOCK_TOKENS = re.compile(r"(.*```py[\w\W]*)|(.*```)")
RGX_PYTABS_DIRECTIVE = re.compile(r"<!-- ?autopytabs: ?(.*)-->")

PyTabDirective = Literal["disable", "enable", "disable-block"]
PYTAB_DIRECTIVES: set[PyTabDirective] = {"disable", "enable", "disable-block"}


def _strip_indentation(lines: list[str]) -> tuple[list[str], str]:
    if not lines:
        return [], ""
    first_line = lines[0]
    indent_char = ""
    if first_line[0] in [" ", "\t"]:
        indent_char = first_line[0]
    indent = indent_char * (len(first_line) - len(first_line.lstrip(indent_char)))
    if indent:
        return [line.split(indent, 1)[1] if line else "" for line in lines], indent
    return lines, indent


def _add_indentation(source: str | list[str], indentation: str) -> str:
    lines = source.splitlines() if isinstance(source, str) else source
    return "\n".join(indentation + line for line in lines)


def _get_pytabs_directive(line: str) -> PyTabDirective | None:
    match = RGX_PYTABS_DIRECTIVE.match(line)
    if match:
        matched_directive = match.group(1).strip()
        if matched_directive in PYTAB_DIRECTIVES:
            return cast(PyTabDirective, matched_directive)
        raise RuntimeError(f"Invalid AutoPytabs directive: {matched_directive!r}")
    return None


def _extract_code_blocks(lines: list[str]) -> tuple[list[str], dict[int, list[str]]]:
    in_block = False
    enabled = True
    new_lines: list[str] = []

    to_transform = {}

    start = 0
    for i, line in enumerate(lines):
        is_comment_line = False
        directive = _get_pytabs_directive(line)
        if directive:
            is_comment_line = True
            if directive == "disable":
                enabled = False
            elif directive == "enable":
                enabled = True

        match = RGX_BLOCK_TOKENS.match(line)
        if match:
            if match.group(1):
                in_block = True
                start = i
            elif match.group(2) and in_block:
                in_block = False
                block = lines[start : i + 1]
                block_directive = _get_pytabs_directive(lines[start - 1])
                if enabled and block_directive != "disable-block":
                    to_transform[len(new_lines)] = block
                    new_lines.append("")
                else:
                    new_lines.extend(block)
            else:
                new_lines.append(line)
        elif not in_block and not is_comment_line:
            new_lines.append(line)

    return new_lines, to_transform


def _build_tab(title: str, body: list[str], selected: bool) -> str:
    out = f'==={"+" if selected else ""} "{title}"\n'
    out += _add_indentation(body, indentation="    ")
    out += "\n"
    return out


def _build_tabs(
    *,
    versioned_code: VersionedCode,
    head: str,
    tail: str,
    tab_title_template: str,
    default_tab_version: VersionTuple,
    reverse_order: bool,
) -> str:
    out = []
    for version, code in versioned_code.items():
        version_string = f"{version[0]}.{version[1]}"
        lines = [head, *code.splitlines(), tail]
        tab_title = tab_title_template.format(min_version=version_string)
        out.append(
            _build_tab(
                title=tab_title, body=lines, selected=version == default_tab_version
            )
        )
    if reverse_order:
        out = reversed(out)  # type: ignore[assignment]
    return "\n".join(out)


def _convert_block(
    *,
    block: list[str],
    versions: list[VersionTuple],
    tab_title_template: str,
    cache: Cache | None,
    default_tab_strategy: Literal["highest", "lowest"],
    reverse_order: bool,
) -> str:
    block, indentation = _strip_indentation(block)
    head, *code_lines, tail = block
    code = "\n".join(code_lines)
    versioned_code = version_code(code, versions, cache=cache)

    if len(versioned_code) > 1:
        versions = list(versioned_code.keys())
        default_tab_version = versions[-1 if default_tab_strategy == "highest" else 0]

        code = _build_tabs(
            versioned_code=versioned_code,
            head=head,
            tail=tail,
            tab_title_template=tab_title_template,
            default_tab_version=default_tab_version,
            reverse_order=reverse_order,
        )
    else:
        code = "\n".join([head, versioned_code[versions[0]], tail])

    code = _add_indentation(code, indentation)

    return code


class UpgradePreprocessor(Preprocessor):
    def __init__(
        self,
        *args: Any,
        min_version: str,
        max_version: str,
        tab_title_template: str | None = None,
        cache: Cache | None = None,
        default_tab_strategy: Literal["highest", "lowest"] = "highest",
        reverse_order: bool = False,
        **kwargs: Any,
    ) -> None:
        self.min_version = VersionTuple.from_string(min_version)
        self.max_version = VersionTuple.from_string(max_version)
        self.versions = get_version_requirements(self.min_version, self.max_version)
        self.tab_title_template = tab_title_template or "Python {min_version}+"
        self.cache = cache
        self.default_tab_strategy = default_tab_strategy
        self.reverse_order = reverse_order
        super().__init__(*args, **kwargs)

    def run(self, lines: list[str]) -> list[str]:
        new_lines, to_transform = _extract_code_blocks(lines)

        output_lines = []
        for i, line in enumerate(new_lines):
            block_to_transform = to_transform.get(i)
            if block_to_transform:
                transformed_block = _convert_block(
                    block=block_to_transform,
                    versions=self.versions,
                    tab_title_template=self.tab_title_template,
                    cache=self.cache,
                    default_tab_strategy=self.default_tab_strategy,
                    reverse_order=self.reverse_order,
                ).splitlines()
                output_lines.extend(transformed_block)
            else:
                output_lines.append(line)

        return output_lines


class AutoPyTabsExtension(Extension):
    def __init__(self, *args: Any, cache: Cache | None, **kwargs: Any):
        self.config = {
            "min_version": ["3.7", "minimum version"],
            "max_version": ["3.11", "maximum version"],
            "tab_title_template": ["", "tab title format-string"],
            "default_tab": ["highest", "version tab to preselect"],
            "reverse_order": [False, "reverse the order of tabs"],
        }
        self.cache = cache
        super().__init__(*args, **kwargs)

    def extendMarkdown(self, md: Markdown) -> None:
        """Register the extension."""
        self.md = md
        md.registerExtension(self)
        config = self.getConfigs()

        md.preprocessors.register(
            UpgradePreprocessor(
                min_version=config["min_version"],
                max_version=config["max_version"],
                tab_title_template=config["tab_title_template"],
                cache=self.cache,
                default_tab_strategy=config["default_tab"],
                reverse_order=config["reverse_order"],
            ),
            "auto_pytabs",
            32,
        )


def makeExtension(**kwargs: Any) -> AutoPyTabsExtension:
    return AutoPyTabsExtension(**kwargs)
